@@ -31,7 +31,7 @@
 
 #include <config.h>
 
-/* For >= 500 for pread/pwrite and readdir_r; >= 700 for utimensat */
+/* For >= 500 for pread/pwrite; >= 700 for utimensat */
 #define _XOPEN_SOURCE 700
 
 /* For flock() on FreeBSD. It otherwise gets hidden by _XOPEN_SOURCE  */
@@ -72,8 +72,14 @@
 #include <sys/xattr.h>
 #endif
 
-#ifdef __LINUX__
+#ifdef __linux__
+#include <sys/mman.h>
+#include <fcntl.h>
 #include <linux/fs.h>  // For BLKGETSIZE64
+
+#ifndef O_DIRECT
+#define O_DIRECT 00040000 /* direct disk access hint */
+#endif
 #endif
 
 #include <fuse.h>
@@ -187,7 +193,12 @@ static struct Settings {
     int enable_lock_forwarding;
 
     int enable_ioctl;
-    
+
+#ifdef __linux__
+    int forward_odirect;
+    size_t odirect_alignment;
+#endif
+
     uid_t uid_offset;
     gid_t gid_offset;
 
@@ -220,6 +231,10 @@ static int apply_gid_offset(gid_t *gid);
 static int unapply_uid_offset(uid_t *uid);
 static int unapply_gid_offset(gid_t *gid);
 
+#ifdef __linux__
+static size_t round_up_buffer_size_for_direct_io(size_t size);
+#endif
+
 /* FUSE callbacks */
 static void *bindfs_init();
 static void bindfs_destroy(void *private_data);
@@ -381,15 +396,15 @@ static int getattr_common(const char *procpath, struct stat *stbuf)
     /* Block files as regular files. */
     if (settings.block_devices_as_files && S_ISBLK(stbuf->st_mode)) {
         stbuf->st_mode ^= S_IFBLK | S_IFREG;  // Flip both bits
-#ifdef __LINUX__
+        int fd = open(procpath, O_RDONLY);
+#ifdef __linux__
         uint64_t size;
-        ioctl(file, BLKGETSIZE64, &size);
+        ioctl(fd, BLKGETSIZE64, &size);
         stbuf->st_size = (off_t)size;
         if (stbuf->st_size < 0) {  // Underflow
             return -EOVERFLOW;
         }
 #else
-        int fd = open(procpath, O_RDONLY);
         if (fd == -1) {
             return -errno;
         }
@@ -585,6 +600,21 @@ static int unapply_gid_offset(gid_t *gid) {
     return 1;
 }
 
+#ifdef __linux__
+static size_t round_up_buffer_size_for_direct_io(size_t size)
+{
+    // `man 2 open` says this should be block-size aligned and that there's no
+    // general way to determine this. Empirically the page size should be good enough.
+    // See also: Pull Request #74
+    size_t alignment = settings.odirect_alignment;
+    size_t rem = size % alignment;
+    if (rem == 0) {
+        return size;
+    }
+    return size - rem + alignment;
+}
+#endif
+
 
 static void *bindfs_init()
 {
@@ -695,17 +725,14 @@ static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
         pc_ret = NAME_MAX;
     }
 
-    struct dirent *de_buf =
-        malloc(offsetof(struct dirent, d_name) + pc_ret + 1);
     int result = 0;
     while (1) {
-        struct dirent *de;
-        result = readdir_r(dp, de_buf, &de);
-        if (result != 0) {
-            result = -result;
-            break;
-        }
+        errno = 0;
+        struct dirent *de = readdir(dp);
         if (de == NULL) {
+            if (errno != 0) {
+                result = -errno;
+            }
             break;
         }
 
@@ -721,14 +748,12 @@ static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
         // consider it an error if it does. It is undocumented whether it sets
         // errno in that case, so we zero it first and set it ourself if it
         // doesn't.
-        errno = 0;
         if (filler(buf, de->d_name, &st, 0) != 0) {
             result = errno != 0 ? -errno : -EIO;
             break;
         }
     }
 
-    free(de_buf);
     closedir(dp);
     return result;
 }
@@ -1060,7 +1085,14 @@ static int bindfs_create(const char *path, mode_t mode, struct fuse_file_info *f
     mode |= S_IFREG; /* tell permchain_apply this is a regular file */
     mode = permchain_apply(settings.create_permchain, mode);
 
-    fd = open(real_path, fi->flags, mode & 0777);
+    int flags = fi->flags;
+#ifdef __linux__
+    if (!settings.forward_odirect) {
+        flags &= ~O_DIRECT;
+    }
+#endif
+
+    fd = open(real_path, flags, mode & 0777);
     if (fd == -1) {
         free(real_path);
         return -errno;
@@ -1083,7 +1115,14 @@ static int bindfs_open(const char *path, struct fuse_file_info *fi)
     if (real_path == NULL)
         return -errno;
 
-    fd = open(real_path, fi->flags);
+    int flags = fi->flags;
+#ifdef __linux__
+    if (!settings.forward_odirect) {
+        flags &= ~O_DIRECT;
+    }
+#endif
+
+    fd = open(real_path, flags);
     free(real_path);
     if (fd == -1)
         return -errno;
@@ -1098,14 +1137,34 @@ static int bindfs_read(const char *path, char *buf, size_t size, off_t offset,
     int res;
     (void) path;
 
+    char *target_buf = buf;
+
     if (settings.read_limiter) {
         rate_limiter_wait(settings.read_limiter, size);
     }
 
-    res = pread(fi->fh, buf, size, offset);
+#ifdef __linux__
+    size_t mmap_size = 0;
+    if ((fi->flags & O_DIRECT) && settings.forward_odirect) {
+        mmap_size = round_up_buffer_size_for_direct_io(size);
+        target_buf = mmap(NULL, mmap_size, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, 0, 0);
+        if (target_buf == MAP_FAILED) {
+            return -ENOMEM;
+        }
+    }
+#endif
+
+    res = pread(fi->fh, target_buf, size, offset);
     if (res == -1)
         res = -errno;
 
+#ifdef __linux__
+    if (target_buf != buf) {
+        memcpy(buf, target_buf, size);
+        munmap(target_buf, mmap_size);
+    }
+#endif
+
     return res;
 }
 
@@ -1114,15 +1173,34 @@ static int bindfs_write(const char *path, const char *buf, size_t size,
 {
     int res;
     (void) path;
+    char *source_buf = (char*)buf;
 
     if (settings.write_limiter) {
         rate_limiter_wait(settings.write_limiter, size);
     }
 
-    res = pwrite(fi->fh, buf, size, offset);
+#ifdef __linux__
+    size_t mmap_size = 0;
+    if ((fi->flags & O_DIRECT) && settings.forward_odirect) {
+        mmap_size = round_up_buffer_size_for_direct_io(size);
+        source_buf = mmap(NULL, mmap_size, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, 0, 0);
+        if (source_buf == MAP_FAILED) {
+            return -ENOMEM;
+        }
+        memcpy(source_buf, buf, size);
+    }
+#endif
+
+    res = pwrite(fi->fh, source_buf, size, offset);
     if (res == -1)
         res = -errno;
 
+#ifdef __linux__
+    if (source_buf != buf) {
+        munmap(source_buf, mmap_size);
+    } 
+#endif
+
     return res;
 }
 
@@ -1493,6 +1571,7 @@ static void print_usage(const char *progname)
            "  --block-devices-as-files  Show block devices as regular files.\n"
            "  --multithreaded           Enable multithreaded mode. See man page\n"
            "                            for security issue with current implementation.\n"
+           "  --forward-odirect=...      Forward O_DIRECT (it's cleared by default).\n"
            "\n"
            "FUSE options:\n"
            "  -o opt[,opt,...]          Mount options.\n"
@@ -1917,6 +1996,7 @@ int main(int argc, char *argv[])
         char *resolved_symlink_deletion;
         int no_allow_other;
         int multithreaded;
+        char *forward_odirect;
         char *uid_offset;
         char *gid_offset;
     } od;
@@ -1988,6 +2068,7 @@ int main(int argc, char *argv[])
         OPT2("--disable-lock-forwarding", "disable-lock-forwarding", OPTKEY_DISABLE_LOCK_FORWARDING),
         OPT2("--enable-ioctl", "enable-ioctl", OPTKEY_ENABLE_IOCTL),
         OPT_OFFSET2("--multithreaded", "multithreaded", multithreaded, -1),
+        OPT_OFFSET2("--forward-odirect=%s", "forward-odirect=%s", forward_odirect, -1),
         OPT_OFFSET2("--uid-offset=%s", "uid-offset=%s", uid_offset, 0),
         OPT_OFFSET2("--gid-offset=%s", "gid-offset=%s", gid_offset, 0),
 
@@ -2038,6 +2119,10 @@ int main(int argc, char *argv[])
     settings.enable_ioctl = 0;
     settings.uid_offset = 0;
     settings.gid_offset = 0;
+#ifdef __linux__
+    settings.forward_odirect = 0;
+    settings.odirect_alignment = 0;
+#endif
 
     atexit(&atexit_func);
 
@@ -2127,7 +2212,7 @@ int main(int argc, char *argv[])
         char* endptr = od.uid_offset;
         settings.uid_offset = strtoul(od.uid_offset, &endptr, 10);
         if (*endptr != '\0') {
-            fprintf(stderr, "Error: Value of --uid-offset must be a positive integer.\n");
+            fprintf(stderr, "Error: Value of --uid-offset must be an integer.\n");
             return 1;
         }
     }
@@ -2144,11 +2229,29 @@ int main(int argc, char *argv[])
         char* endptr = od.gid_offset;
         settings.gid_offset = strtoul(od.gid_offset, &endptr, 10);
         if (*endptr != '\0') {
-            fprintf(stderr, "Error: Value of --gid-offset must be a positive integer.\n");
+            fprintf(stderr, "Error: Value of --gid-offset must be an integer.\n");
             return 1;
         }
     }
 
+    if (od.forward_odirect) {
+#ifdef __linux__
+        settings.forward_odirect = 1;
+        char* endptr = od.forward_odirect;
+        settings.odirect_alignment = strtoul(od.forward_odirect, &endptr, 10);
+        if (*endptr != '\0') {
+            fprintf(stderr, "Error: Value of --forward-odirect must be an integer.\n");
+            return 1;
+        }
+        if (settings.odirect_alignment == 0) {
+            fprintf(stderr, "Error: Value of --forward-odirect must be positive.\n");
+            return 1;
+        }
+#else
+        fprintf(stderr, "Warning: --forward-odirect is not supported on this platform.\n");
+#endif
+    }
+
     /* Parse user and group for new creates */
     if (od.create_for_user) {
         if (getuid() != 0) {
