@@ -763,15 +763,56 @@ testenv("", :title => "mntdir with newline", :mntdir_name => "a\nb") do
 end
 
 # Pull Request #73
-if `uname`.strip == 'Linux' && `which unshare` != ''
+if `uname`.strip == 'Linux' &&
+    `uname -r`.strip =~ /^[456789]/ &&  # 3.x kernels used by CentOS 7 and older don't support all `unshare` options despite the userspace binary supporting them
+    `which unshare` != '' &&
+    `unshare --help`.include?("--map-root-user") &&
+    `unshare --help`.include?("--user")
   root_testenv("--gid-offset=10000", :title => "setgid and gid-offset") do
     system("chmod g+s src")
-    system("unshare --map-root-user --user #{$nobody} mkdir mnt/dir")
+    system("unshare --map-root-user --user mkdir mnt/dir")
     assert { File.stat("src/dir").gid == 0 }
     assert { File.stat("mnt/dir").gid == 10000 }
   end
 end
 
+# Pull Request #74
+if `uname`.strip == 'Linux'
+  def odirect_data
+    ('abc' * 10000)[0...8192]
+  end
+
+  testenv("", :title => "O_DIRECT reads with O_DIRECT ignored") do
+    File.write("src/f", odirect_data)
+    read_data = `#{$tests_dir}/odirect_read mnt/f`
+    assert { $?.success? }
+    assert { read_data == odirect_data }
+  end
+
+  testenv("", :title => "O_DIRECT writes with O_DIRECT ignored") do
+    IO.popen("#{$tests_dir}/odirect_write mnt/f", "w") do |pipe|
+      pipe.write(odirect_data)
+    end
+    assert { $?.success? }
+    assert { File.read("src/f") == odirect_data }
+  end
+
+  testenv("--forward-odirect=512", :title => "O_DIRECT reads with O_DIRECT forwarded") do
+    File.write("src/f", odirect_data)
+    read_data = `#{$tests_dir}/odirect_read mnt/f`
+    assert { $?.success? }
+    assert { read_data == odirect_data }
+  end
+
+  testenv("--forward-odirect=512", :title => "O_DIRECT writes with O_DIRECT forwarded") do
+    IO.popen("#{$tests_dir}/odirect_write mnt/f", "w") do |pipe|
+      pipe.write(odirect_data)
+    end
+    assert { $?.success? }
+    assert { File.read("src/f") == odirect_data }
+  end
+end
+
 if `uname`.strip != 'FreeBSD'  # -o dev is not supported on FreeBSD
   root_testenv("-odev") do
     system("mknod mnt/zero c 1 5")
